#ifndef FINDSAMPLES_H
#define FINDSAMPLES_H

/** \author Sebastian Lehmann */

#include "log2.h"


//----------------------------------------------------------------------------------------------------------------------

// HELPERS

// e.g.:
// k = 2 => 1
// k = 4 => 2 1 3
// k = 8 => 4 2 6 1 3 5 7
template<int K>
void makeTreeIndices(int * tree, int factor = 1)
{
    makeTreeIndices<K/2>(tree, factor*2);
    for(int i = 0; i < K/2; ++i)
        tree[i + K/2-1] = (i * 2 + 1) * factor;
}
template<>
void makeTreeIndices<1>(int * tree, int factor)
{
    tree[0] = factor;
}

// e.g.:
// k = 2 => ø 1
// k = 4 => ø 2 1 3
// k = 8 => ø 4 2 6 1 3 5 7
template<int K>
int *treeIndices()
{
    static int cache[K];
    cache[0] = -42; // don't touch this!
    static bool initialized = false;
    if(!initialized)
        makeTreeIndices<K>(cache + 1);
    return cache;
}


/** Swaps a with b if not a < b. */
template<class T>
inline void conditionalSwap(T &a, T &b)
{
    std::swap(a, a < b ? a : b);
}



//----------------------------------------------------------------------------------------------------------------------

// as in lecture "Algorithm Engineering" by Sanders

template<class T, int K, int ALPHA, class It /* deduced */>
void findSamplesMultiRandomStdSort(It begin, It end, It target)
{
    static const int NUM_SAMPLES = K * ALPHA;
    int n = (end - begin);

    // Pick samples (alpha * k)
    T samples[NUM_SAMPLES];
    for (int i = 0; i < NUM_SAMPLES; ++i)
        samples[i] = *(begin + (rand() % n));

    // Sort samples
    std::sort(samples, samples + NUM_SAMPLES);

    // Find quantiles
    for(int i = 1; i < K; ++i)
        *(target + i - 1) = samples[treeIndices<K>()[i] * ALPHA - 1];
}


//----------------------------------------------------------------------------------------------------------------------

// slightly changed - replaced std::sort with optimized merge sort

// Merger merges COUNT arrays of sizes SIZE from source to target iterator
template<class T, class It, int COUNT, int SIZE>
struct Merger
{
    void operator()(It source, It target)
    {
        Merger<T, It, (COUNT * 2), (SIZE / 2)>()(target, source);
        // merge buffer pairs
        for(int b = 0; b < COUNT; ++b)
        {
            It s1 = source + (b * SIZE);
            It s2 = source + (b * SIZE) + (SIZE / 2);
            It s1end = s2;
            It s2end = source + (b * SIZE) + SIZE;

            It *sx [] = { &s1, &s2 };

            It t = target + (b * SIZE);
            for(; (s1 < s1end) && (s2 < s2end); ++t)
            {
                bool x = *s2 < *s1; // take from right buffer?
                *t = **sx[x];       // take from either s1 or s2, depending on x
                s1 += (1 - x);
                s2 += x;
            }
            for(; (s1 < s1end); ++t)
                *t = *(s1++);
            for(; (s2 < s2end); ++t)
                *t = *(s2++);
        }
    }
};
// Base case: merge 2x2 => 4 elements
template<class T, class It, int COUNT>
struct Merger<T, It, COUNT, 2>
{
    void operator()(It source, It target)
    {
        if(LOG2<COUNT>::value % 2)
            std::copy(source, source + COUNT, target);
        for(int b = 0; b < COUNT; ++b)
        {
            conditionalSwap(*(target), *(target + 1));
            target += 2;
        }
    }
};
// Merge sort
template<class T, int SIZE, class It /* deduced */>
void mergeSort(It begin) {
    static T tmp[SIZE]; // temporary buffer
    Merger<T, It, 1, SIZE>()(tmp, begin);
}


template<class T, int K, int ALPHA, class It /* deduced */>
void findSamplesMultiRandomMergeSort(It begin, It end, It target)
{
    static const int NUM_SAMPLES = K * ALPHA;
    int n = (end - begin);

    // Pick samples (alpha * k)
    T samples[NUM_SAMPLES];
    for (int i = 0; i < NUM_SAMPLES; ++i)
        samples[i] = *(begin + (rand() % n));

    // Sort samples
    //std::sort(samples, samples + NUM_SAMPLES);
    mergeSort<T, NUM_SAMPLES>(samples);

    // Find quantiles
    for(int i = 1; i < K; ++i)
        *(target + i - 1) = samples[treeIndices<K>()[i] * ALPHA - 1];
}



//----------------------------------------------------------------------------------------------------------------------


// Median-of-medians (https://en.wikipedia.org/wiki/Selection_algorithm#Linear_general_selection_algorithm_-_Median_of_Medians_algorithm)
template<class It>
It median1(It begin) { return begin; }
template<class It>
It median2(It begin) { return begin + (rand() % 2); }
template<class It>
It median3(It begin) {
    std::sort(begin, begin + 3);
    return begin + 1;
}
template<class It>
It median4(It begin) {
    std::sort(begin, begin + 4);
    return begin + 1 + (rand() % 2);
}
template<class It>
It median5(It begin) {
    std::sort(begin, begin + 5);
    return begin + 2;
}
//...
template<class T, class It /* deduced */>
T findSampleMom(It begin, It end)
{
    static const int BASE_CASE = 16;
    int n = (end - begin);

    // base case:
    if(n <= BASE_CASE)
        return *(begin + (rand() % n));

    int mediansCount = n / 5;
    T *medians = new T[mediansCount];
    int i;
    for(i = 0; i < mediansCount; ++i)
        medians[i] = *median5(begin + (i * 5));

    T m = findSampleMom<T>(medians, medians + mediansCount);
    delete [] medians;
    return m;
}



#endif // FINDSAMPLES_H
